/*
 * Copyright (c) 2000-2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 ** Anthony Mallet - Tue Aug  8 2000
 ** Matthieu Herrb - January 2006
 **
 **/

#include "config.h"

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#ifndef NO_THREADS
#include <pthread.h>
#endif
#include <signal.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <portLib.h>
#include <posterLib.h>
#include <h2errorLib.h>
#include <h2initGlob.h>
#include <taskLib.h>

#include "picoweb.h"

static int stop;

#ifndef RTLD_DEFAULT
static void *webHandle;
#endif /* RTLD_DEFAULT */

/* Parameter structure for CGI handler */
struct CGI {
	CGIFUNCPTR fptr;	/* actual function to be called */
	int fd;			/* file descriptor of client */
	int argc;
	char **argv;
	char **argn;
};

/* Static functions */
static int webTask(int);
static int getLine(int, char *, int);
static void invokeCgi(int, char *);
static void cgiWrapper(void (*)(FILE *, int, char **, char **),
    int, int, char **, char **);
static void *cgiWrapper2(void *);
static int splitArgs(char *, char ***, char ***);

/*----------------------------------------------------------------------*/

static void
usage(char *progname)
{
	fprintf(stderr, "usage: %s [-d][-h][-p port][-v]\n", progname);
	fprintf(stderr, "\t-d debug mode: do not daemonize\n");
	fprintf(stderr, "\t-h help: print this help message and exit\n");
	fprintf(stderr,
	       "\t-p <port> : specify the TCP port used (default 8080)\n");
	fprintf(stderr, "\t-v version: print picoweb version and exit\n");
}

/*----------------------------------------------------------------------*/

int
main(int argc, char *argv[])
{
	int ch, debugFlag, port;
	char buf[80];

	debugFlag = 0;
	port = 8080;
	while ((ch = getopt(argc, argv, "dhp:v")) != -1) {
		switch (ch) {
		case 'd':
			debugFlag = 1;
			break;
		case 'h':
			usage(argv[0]);
			exit(0);
		case 'p':
			port = atoi(optarg);
			break;
		case 'v':
			printf("%s version %s\n", PACKAGE, VERSION);
			exit(0);
		default:
			usage(argv[0]);
			exit(2);
		}
	}
	if (!debugFlag) {
		if (daemon(0,1) < 0) {
			fprintf(stderr, "picoweb: %s\n", strerror(errno));
			exit(2);
		}
	}
	if (h2initGlob(0) != OK) {
		fprintf(stderr, "picoweb: %s\n", 
		    h2getErrMsg(errno, buf, sizeof(buf)));
		return EXIT_FAILURE;
	}

	if (webTask(port) == ERROR) {
		return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}

/*----------------------------------------------------------------------*/
static void
sighandler(int sig)
{
	stop = 1;
}

/*
 * Main server loop -----------------------------------------------------
 *
 */

static int
webTask(int port)
{
	int s, fd;
	fd_set fdSet;
	static char buf[512];
	static char cgi[512];
	struct sockaddr_in addr;
	socklen_t sockLen = sizeof(addr);
	int reuseSocket = 1;

#ifndef RTLD_DEFAULT
	webHandle = dlopen(NULL, RTLD_LAZY);
#endif /* RTLD_DEFAULT */

	/* create socket */
	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s <= 0) {
		perror("web:socket");
		return ERROR;
	}

	/* set options */
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
		(char *)&reuseSocket, sizeof(reuseSocket)) < 0) {
		perror("web:setsockopt");
		close(s);
		return ERROR;
	}
	/* bind socket */
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr= htonl(INADDR_ANY);
	addr.sin_port= htons(port);

	if (bind(s, (struct sockaddr *)&addr, sizeof(struct sockaddr_in))) {
		perror("web:bind");
		close(s);
		return ERROR;
	}

	/* listen */
	if (listen(s, 10)){
		perror("web:listen");
		close(s);
		return ERROR;
	}

	/* install signal handler */
	signal(SIGHUP, sighandler);
	signal(SIGPIPE, SIG_IGN);

	/* wait for connections */
	fprintf(stderr, "webTask started on port %d\n", port);

	while(!stop) {
		FD_ZERO(&fdSet);
		FD_SET(s, &fdSet);

		if (select(s+1, &fdSet, NULL, NULL, NULL) < 0) {
			if (errno == EINTR) continue;

			perror("web:select");
			break;
		}

		if (FD_ISSET(s, &fdSet)) {
			errno = 0;
			do {
				fd = accept(s, (struct sockaddr *)&addr,
				    &sockLen);
			} while(fd < 0 && errno == EINTR);
			if (fd < 0) { perror("web:accept"); continue; }

			/* parse html request */
			if (getLine(fd, buf, sizeof(buf)) < 0) {
				perror("web:fgets");
				close(fd);
			} else {
				if (sscanf(buf, "GET %s", cgi)) {
					/* invoke script */
					while(getLine(fd, buf, sizeof(buf))) {
						if (*buf == '\n'
						    || *buf == '\r') break;
					}
					invokeCgi(fd, cgi);
				} else {
					while(getLine(fd, buf, sizeof(buf))) {
						fputs(buf, stderr);
						if (*buf == '\n'
						    || *buf == '\r') break;
					}
					webError(buf, 501, fd);
				}
			}
		}
	}

	/* leave */
	fputs("webTask stopped\n", stderr);

	close(s);

#ifndef RTLD_DEFAULT
	dlclose(webHandle);
#endif

	return OK;
}

static int
getLine(int fd, char *buf, int maxLen)
{
	int a;
	char c=0;
	int r=0, i = maxLen;
	errno = 0;
	while(i>0 && c!= '\n' && !errno){
		a=read(fd, buf, 1);
		if (a == 1) {
			i--;
			r++;
			c= *(buf++);
		}
		else if (a==0) break;
		if (errno == EINTR) errno = 0;
	}
	*buf=0;
	return r;
}


/*
 * invokeCgi ------------------------------------------------------------
 *
 * Parse an url and invoke the cgi script attached to it
 */

static void
invokeCgi(int fd, char *cgi)
{
	struct CGI *arg;
#ifndef NO_THREADS
	long tid;
#endif
	CGIFUNCPTR fptr = NULL;
	int argc;
	int i;
	char **argv, **argn;
	struct timeval tv;
	char buf[26], *date;

	/* parse request */
	if (*cgi != '/') {
		webError(cgi, 400, fd);
		return;
	}
	cgi++;
	argc = splitArgs(cgi, &argv, &argn);
	if (argc <= 0) {
		webError(cgi, 400, fd);
		return;
	}
	gettimeofday(&tv, NULL);
	date = ctime_r(&tv.tv_sec, buf);
	date[strlen(date) - 1] = '\0';	/* remove final newline */
	fprintf(stderr, "%s: picoweb: request: %s", date, argv[0]);
	for(i=1;i<argc;i++) {
		if (argn[i] == NULL)
			fprintf(stderr, ", %s", argv[i]);
		else
			fprintf(stderr, ", %s=%s", argn[i], argv[i]);
	}
	fprintf(stderr, "\n");

	/* look for symbol */
	fptr = autoload(argv[0]);

	if(fptr == NULL) {
#ifdef DEBUG
		fprintf(stderr, "picoweb: %s ... not found\n", argv[0]);
#endif
		webError(argv[0], 404, fd);
		return;
	}

	/* spawn cgi */
	arg = (struct CGI *)malloc(sizeof(struct CGI));
	arg->fptr = fptr;
	arg->fd = fd;
	arg->argc = argc;
	arg->argv = argv;
	arg->argn = argn;

#ifndef NO_THREADS
	tid = taskSpawn2(NULL, 100, VX_FP_TASK, 65536, cgiWrapper2, arg);
	if (tid == ERROR) {
		h2perror("taskSpawn");
		return;
	}
#else
	cgiWrapper(fptr, fd, argc, (int)argv, (int)argn);
#endif

#ifdef DEBUG
	fprintf(stderr, "... served\n");
#endif
}


/*
 * cgiWrapper -----------------------------------------------------------
 *
 * Invoke cgi
 */

static void
cgiWrapper(void (*fptr)(FILE *f, int argc, char **argv, char **argn),
    int fd, int argc, char *argv[], char *argn[])
{
	FILE *f;

	f = fdopen(fd, "w");
	if (f == NULL) {
		fprintf(stderr, "picoweb: %s: %s\n",
		    __func__, strerror(errno));
		free(argv[0]); free(argv);
		close(fd);
		return;
	}

	(*fptr)(f, argc, argv, argn);
#ifdef DEBUG
	fprintf(stderr, "web report: stopped %s\n", argv[0]);
#endif
	fclose(f);
	free(argv[0]); free(argv);
	return;
}

static void *
cgiWrapper2(void *arg)
{
	struct CGI *cgi = (struct CGI *)arg;
	cgiWrapper(cgi->fptr, cgi->fd, cgi->argc, cgi->argv, cgi->argn);
	return NULL;
}


/*
 * splitArgs ------------------------------------------------------------
 *
 * split args into an argv[argc] array.
 */

static int
splitArgs(char *args, char ***argv, char ***argn)
{
	int i, argc;
	char *arg, *ppLast;
	size_t s;

	/* copy args */
	s = strlen(args)+1;
	arg = malloc(s);
	if (arg == NULL) return -1;
	strlcpy(arg, args, s);
	args = arg;

	/* parse %2 strings */
	for(arg=strstr(args,"%2");arg!=NULL;arg=strstr(arg+1,"%2")) {
		i = 0;
		switch(arg[2]) {
		case 'f': *arg='/'; i = 1; break;
		case 'd': *arg='-'; i = 1; break;
		}
		if (i>0) {
			memmove(arg+i, arg+3, strlen(arg+3)+1);
		}
	}

	/* count args */
	for(argc = 0, arg = strtok_r(args, "&?", &ppLast);
	    arg != NULL;
	    argc++, arg = strtok_r(NULL, "&?", &ppLast));
	if (argc == 0) return 0; /* should never happen */

	/* alloc argv and argn array */
	*argv = malloc(argc * sizeof(**argv));
	if (*argv == NULL) return -1;
	*argn = malloc(argc * sizeof(**argn));
	if (*argn == NULL) { free(*argv); return -1; }

	/* fill array */
	for(i=0, arg=args; i<argc; arg += strlen(arg)+1, i++)
		(*argv)[i] = arg;

	/* split args into (name, value) pairs */
	for(i=0; i<argc; i++) {
		(*argn)[i] = (*argv)[i];
		if ((arg = strchr((*argv)[i], '=')) != NULL) {
			*arg++ = 0;
			(*argv)[i] = arg;
		}
	}

	return argc;
}


/*
 * errors ---------------------------------------------------------------
 */

int
webError(char *text, int n, int fd)
{
	char buf[64];

	switch(n) {
	case 400:
		strlcpy(buf, "HTTP/1.0 400 Bad Request\n\n", sizeof(buf));
		break;

	case 404:
		strlcpy(buf, "HTTP/1.0 404 Not Found\n\n", sizeof(buf));
		break;

	case 500:
		strlcpy(buf, "HTTP/1.0 500 Internal Server Error\n\n",
			sizeof(buf));
		break;
	case 501:
		strlcpy(buf, "HTTP/1.0 501 Not Implemented\n\n", sizeof(buf));
		break;
	case 503:
		strlcpy(buf, "HTTP/1.0 503 Service Unavailable\n\n",
			sizeof(buf));
		break;
	}

	write(fd, buf, strlen(buf));
	close(fd);
	fprintf(stderr, "picoweb: web report: %s %s\n", text, buf);
	return ERROR;
}
